'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

function _interopNamespace(e) {
  if (e && e.__esModule) { return e; } else {
    var n = {};
    if (e) {
      Object.keys(e).forEach(function (k) {
        var d = Object.getOwnPropertyDescriptor(e, k);
        Object.defineProperty(n, k, d.get ? d : {
          enumerable: true,
          get: function () {
            return e[k];
          }
        });
      });
    }
    n['default'] = e;
    return n;
  }
}

var typedi = require('typedi');
var tslib = require('tslib');
var signale = _interopDefault(require('signale'));
var path = require('path');
var path__default = _interopDefault(path);
var fs = require('fs');
var glob = _interopDefault(require('glob'));
var express = _interopDefault(require('express'));
var os = _interopDefault(require('os'));
var chokidar = _interopDefault(require('chokidar'));

(function (PluginOrderEnum) {
    PluginOrderEnum["BEFORE_API_INIT"] = "BEFORE_API_INIT";
    PluginOrderEnum["API_INIT"] = "API_INIT";
    PluginOrderEnum["AFTER_API_INIT"] = "AFTER_API_INIT";
})(exports.PluginOrderEnum || (exports.PluginOrderEnum = {}));
var InternalPluginOrderEnum;
(function (InternalPluginOrderEnum) {
    InternalPluginOrderEnum["FIRST_STAGE"] = "FIRST_STAGE";
    InternalPluginOrderEnum["FINAL_STAGE"] = "FINAL_STAGE";
})(InternalPluginOrderEnum || (InternalPluginOrderEnum = {}));

function isNotEmpty(obj) {
    return obj !== undefined && obj !== null;
}
function isEmpty(obj) {
    return obj === undefined || obj === null;
}
function isString(obj) {
    return Object.prototype.toString.call(obj) === '[object String]';
}
function isArray(obj) {
    return isNotEmpty(obj) && Array.isArray(obj);
}
function urlJoin(...urls) {
    return ['/', ...urls]
        .join('/')
        .replace(/\/+/g, '/')
        .replace(/.*http:\/*/, 'http://')
        .replace(/.*https:\/*/, 'https://');
}

function cwd() {
    return process.cwd();
}
function getProjectBaseRoot() {
    if (process.env.NODE_ENV === 'development') {
        return path.resolve(cwd(), 'src');
    }
    if (fs.existsSync(path.resolve(cwd(), 'dist', 'src'))) {
        return path.resolve(cwd(), 'dist', 'src');
    }
    return path.resolve(cwd(), 'dist');
}
const ProcessCtrl = {
    stop() {
        if (isEmpty(process.send)) {
            return;
        }
        process.send({
            type: ProcessSingle.STOP
        });
    },
    restart(msg) {
        if (isEmpty(process.send)) {
            return;
        }
        process.send({
            type: ProcessSingle.RESTART,
            data: msg
        });
    }
};
var ProcessSingle;
(function (ProcessSingle) {
    ProcessSingle["STOP"] = "STOP";
    ProcessSingle["RESTART"] = "RESTART";
})(ProcessSingle || (ProcessSingle = {}));

var cors = () => {
    return {
        order: InternalPluginOrderEnum.FIRST_STAGE,
        configHandler(config) {
            if (isEmpty(config) || isEmpty(config.cors)) {
                return {
                    cors: true
                };
            }
            return {
                cors: config.cors
            };
        },
        pluginHandler(app, config) {
            app.use('*', (req, res, next) => {
                if (!req.get('Origin')) {
                    return next();
                }
                res.set('Access-Control-Allow-Origin', req.headers.origin);
                res.set('Access-Control-Allow-Methods', 'POST, GET, OPTIONS, PATCH, PUT, HEAD, TRACE, DELETE');
                res.set('Access-Control-Allow-Headers', 'X-Requested-With, Content-Type, Accept, Range');
                res.set('Access-Control-Allow-Credentials', 'true');
                if ('OPTIONS' === req.method) {
                    return res.status(200).end();
                }
                return next();
            });
        }
    };
};

const defaultSuccessResponseResolver = (data) => {
    return {
        code: 200,
        data
    };
};
const defaultFailureResponseResolver = (error) => {
    return {
        code: error.code,
        message: error.message || 'Internal error'
    };
};
var api = () => {
    return {
        order: exports.PluginOrderEnum.API_INIT,
        configHandler(config) {
            const conf = {
                api: {
                    scanDir: path.resolve(getProjectBaseRoot(), 'controllers'),
                    prefix: '/apis',
                    successResponseResolver: defaultSuccessResponseResolver,
                    failureResponseResolver: defaultFailureResponseResolver
                }
            };
            if (isEmpty(config) || isEmpty(config.api)) {
                return conf;
            }
            if (isNotEmpty(conf.api) && isNotEmpty(config.api.prefix)) {
                conf.api.prefix = config.api.prefix;
            }
            if (isNotEmpty(conf.api) && isNotEmpty(config.api.successResponseResolver)) {
                conf.api.successResponseResolver = config.api.successResponseResolver;
            }
            if (isNotEmpty(conf.api) && isNotEmpty(config.api.failureResponseResolver)) {
                conf.api.failureResponseResolver = config.api.failureResponseResolver;
            }
            return conf;
        },
        pluginHandler(app, config) {
            if (isEmpty(config.api)) {
                return;
            }
            const ext = process.env.NODE_ENV === 'development' ? '.ts' : '.js';
            const controllerFiles = glob.sync(`**/*${ext}`, {
                cwd: config.api.scanDir,
                absolute: true
            });
            return Promise.all(controllerFiles.map(f => new Promise(function (resolve) { resolve(_interopNamespace(require(f))); })));
        }
    };
};

var server = () => {
    return {
        order: InternalPluginOrderEnum.FINAL_STAGE,
        configHandler(config) {
            if (isEmpty(config) || isEmpty(config.server)) {
                return {
                    server: {
                        port: 8080,
                        staticDir: path__default.resolve(cwd(), 'public')
                    }
                };
            }
            return {
                server: {
                    port: isEmpty(config.server.port) ? 8080 : config.server.port,
                    staticDir: isEmpty(config.server.staticDir) ? path__default.resolve(cwd(), 'public') : config.server.staticDir
                }
            };
        },
        pluginHandler(app, config) {
            if (isEmpty(config) ||
                isEmpty(config.server) ||
                isEmpty(config.server.port) ||
                isEmpty(config.server.staticDir)) {
                signale.error('No server config found');
                return;
            }
            const { port, staticDir } = config.server;
            if (isString(staticDir)) {
                app.use(express.static(staticDir));
            }
            return new Promise(resolve => {
                app.listen(port, '0.0.0.0', err => {
                    if (err) {
                        signale.error(err);
                        ProcessCtrl.stop();
                        return;
                    }
                    const addresses = getIpAddresses();
                    signale.success(`App running at below link${addresses.length > 1 ? 's' : ''}:`);
                    addresses.forEach(ip => {
                        signale.info(`http://${ip}:${port}${config['api']['prefix']}`);
                        resolve();
                    });
                });
            });
        }
    };
};
function getIpAddresses() {
    const ifaces = os.networkInterfaces();
    return Object.values(ifaces)
        .reduce((p, c) => p.concat(c), [])
        .filter(iface => iface.family === 'IPv4')
        .map(iface => iface.address);
}

function debounce(func, waitMilliseconds = 50, options = {
    isImmediate: false
}) {
    let timeoutId;
    // tslint:disable: no-function-expression
    return function (...args) {
        const context = this;
        const doLater = function () {
            timeoutId = undefined;
            if (!options.isImmediate) {
                func.apply(context, args);
            }
        };
        const shouldCallNow = options.isImmediate && timeoutId === undefined;
        if (timeoutId !== undefined) {
            clearTimeout(timeoutId);
        }
        timeoutId = setTimeout(doLater, waitMilliseconds);
        if (shouldCallNow) {
            func.apply(context, args);
        }
    };
}

var watcher = () => {
    return {
        order: exports.PluginOrderEnum.BEFORE_API_INIT,
        configHandler(config) {
            return config;
        },
        pluginHandler(app, config) {
            if (process.env.NODE_ENV !== 'development') {
                return;
            }
            const srcDir = path.resolve(cwd(), 'src');
            const debounceRestart = debounce((msg) => {
                ProcessCtrl.restart(msg);
            }, 800);
            chokidar
                .watch(srcDir, { ignoreInitial: true })
                .on('add', path => debounceRestart(`${path} added`))
                .on('addDir', path => debounceRestart(`Dir ${path} added`))
                .on('change', path => debounceRestart(`${path} changed`))
                .on('unlink', path => debounceRestart(`${path} removed`))
                .on('unlinkDir', path => debounceRestart(`Dir ${path} removed`));
        }
    };
};

function getRawUserConfig() {
    try {
        // tslint:disable-next-line: non-literal-require
        const mod = require(path.join(cwd(), '.bootrc.ts'));
        return mod.default || mod;
    }
    catch (error) {
        signale.warn('.bootrc.ts not found, preset config used');
        return {};
    }
}

const rawConfig = getRawUserConfig();
const internalPlugins = [cors(), watcher(), api(), server()];
function getExternalPlugins() {
    const factories = getExternalPluginFactories();
    return factories.map(factory => {
        return factory.module(factory.options);
    });
}
function getExternalPluginFactories() {
    if (!isArray(rawConfig.plugins)) {
        return [];
    }
    if (hasIncorrectPlugin(rawConfig.plugins)) {
        signale.error('incorrect plugin configured');
        ProcessCtrl.stop();
        return [];
    }
    return getExternalPluginModules(rawConfig.plugins);
}
function hasIncorrectPlugin(plugins) {
    return plugins.some(plugin => isEmpty(plugin) || isEmpty(plugin.name));
}
function getExternalPluginModules(plugins) {
    const modulePaths = [
        ...plugins.map(plugin => ({ mp: path.join(cwd(), 'plugins', plugin.name), options: plugin.options })),
        ...plugins.map(plugin => ({ mp: path.join(cwd(), 'node_modules', plugin.name), options: plugin.options }))
    ];
    return modulePaths
        .map(mp => {
        try {
            // tslint:disable-next-line: non-literal-require
            const mod = require(mp.mp);
            return {
                module: mod.default || mod,
                options: mp.options
            };
        }
        catch (error) {
            return null;
        }
    })
        .filter((mod) => !!mod);
}
const plugins = [...internalPlugins, ...getExternalPlugins()];
const userConfig = plugins.reduce((p, c) => {
    return Object.assign(Object.assign({}, p), c.configHandler(p));
}, Object.assign({}, rawConfig));
function execPlugins(order, app) {
    return tslib.__awaiter(this, void 0, void 0, function* () {
        const getPlugins = plugins.filter(p => p.order === order);
        yield Promise.all(getPlugins.map(p => p.pluginHandler(app, Object.assign({}, userConfig)) || Promise.resolve()));
        return undefined;
    });
}
const pluginRunner = {
    firstStage(app) {
        return tslib.__awaiter(this, void 0, void 0, function* () {
            return execPlugins(InternalPluginOrderEnum.FIRST_STAGE, app);
        });
    },
    beforeApiInit(app) {
        return tslib.__awaiter(this, void 0, void 0, function* () {
            return execPlugins(exports.PluginOrderEnum.BEFORE_API_INIT, app);
        });
    },
    apiInit(app) {
        return tslib.__awaiter(this, void 0, void 0, function* () {
            return execPlugins(exports.PluginOrderEnum.API_INIT, app);
        });
    },
    afterApiInit(app) {
        return tslib.__awaiter(this, void 0, void 0, function* () {
            return execPlugins(exports.PluginOrderEnum.AFTER_API_INIT, app);
        });
    },
    lastStage(app) {
        return tslib.__awaiter(this, void 0, void 0, function* () {
            return execPlugins(InternalPluginOrderEnum.FINAL_STAGE, app);
        });
    }
};

const PATH_SETS = {
    get: new Set(),
    post: new Set(),
    delete: new Set(),
    update: new Set(),
    patch: new Set()
};
const CONTROLLER_INSTANCES = new Map();
const HANDLER_PARAMETER_INJECTORS = new Map();
const ExpressToken = new typedi.Token('global.express');
const RestControllerToken = new typedi.Token('controllers');
function setExpressApp(app) {
    typedi.Container.set(ExpressToken, app);
}
function RestController() {
    return (target) => {
        return typedi.Service({ id: RestControllerToken, multiple: true })(target);
    };
}
function GetMapping(path) {
    return createMapping(path, 'get');
}
function PostMapping(path) {
    return createMapping(path, 'post');
}
function DeleteMapping(path) {
    return createMapping(path, 'delete');
}
function UpdateMapping(path) {
    return createMapping(path, 'update');
}
function PatchMapping(path) {
    return createMapping(path, 'patch');
}
function createMapping(path, httpMethod) {
    // tslint:disable-next-line: no-function-expression
    return function (target, propertyKey, descriptor) {
        if (PATH_SETS[httpMethod].has(path)) {
            signale.error(`${httpMethod} ${path} was registered multiple times, please verify your code first`);
            ProcessCtrl.stop();
        }
        injectParameterDecorator(target, propertyKey, descriptor);
        registerAPI(target, propertyKey, path, httpMethod);
        return descriptor;
    };
}
function injectParameterDecorator(target, propertyKey, descriptor) {
    const originalMethod = descriptor.value;
    descriptor.value = function (req, res) {
        const injector = HANDLER_PARAMETER_INJECTORS.get(target);
        if (isEmpty(injector)) {
            return originalMethod.apply(this, [req, res]);
        }
        const handlers = injector[propertyKey];
        if (isEmpty(handlers)) {
            return originalMethod.apply(this, [req, res]);
        }
        const finalArgs = [];
        handlers.forEach((handler, i) => {
            if (isEmpty(handler)) {
                if (i === 0) {
                    return finalArgs.push(req);
                }
                if (i === 1) {
                    return finalArgs.push(res);
                }
                // no valid handler found
                return finalArgs.push(null);
            }
            finalArgs.push(handler(req, res));
        });
        return originalMethod.apply(this, finalArgs);
    };
    return descriptor;
}
function registerAPI(target, propertyKey, path, httpMethod) {
    const app = typedi.Container.get(ExpressToken);
    const { successResponseResolver, failureResponseResolver, prefix } = userConfig.api;
    app[httpMethod](urlJoin(prefix, path), (req, res) => tslib.__awaiter(this, void 0, void 0, function* () {
        try {
            let instance = CONTROLLER_INSTANCES.get(target);
            if (isEmpty(instance)) {
                instance = typedi.Container.getMany(RestControllerToken).find(c => Object.getPrototypeOf(c) === target);
                CONTROLLER_INSTANCES.set(target, instance);
            }
            const result = yield instance[propertyKey](req, res);
            res.json(successResponseResolver(result));
        }
        catch (error) {
            signale.error(error);
            res.json(failureResponseResolver(error));
        }
    }));
}
function createParameterDecorator(handler) {
    // tslint:disable-next-line: no-function-expression
    return function (target, methodName, paramIndex) {
        let injector = HANDLER_PARAMETER_INJECTORS.get(target);
        if (isEmpty(injector)) {
            injector = {};
            HANDLER_PARAMETER_INJECTORS.set(target, injector);
        }
        if (isEmpty(injector[methodName])) {
            // tslint:disable-next-line: prefer-array-literal
            injector[methodName] = new Array(target[methodName].length);
            injector[methodName].fill(null);
        }
        injector[methodName][paramIndex] = handler;
    };
}
function RequestQuery() {
    return createParameterDecorator((req) => req.query);
}
function RequestBody() {
    return createParameterDecorator((req) => req.body);
}
function RequestPathVariable() {
    return createParameterDecorator((req) => req.params);
}
function Request() {
    return createParameterDecorator((req) => req);
}
function Response() {
    return createParameterDecorator((req, res) => res);
}

class BizError extends Error {
    constructor(code, message) {
        super(message);
        this.code = code;
    }
}

const ___internal = {
    pluginRunner,
    setExpressApp,
    ExpressToken,
    ProcessSingle
};

Object.defineProperty(exports, 'Container', {
  enumerable: true,
  get: function () {
    return typedi.Container;
  }
});
Object.defineProperty(exports, 'Inject', {
  enumerable: true,
  get: function () {
    return typedi.Inject;
  }
});
Object.defineProperty(exports, 'Service', {
  enumerable: true,
  get: function () {
    return typedi.Service;
  }
});
Object.defineProperty(exports, 'Token', {
  enumerable: true,
  get: function () {
    return typedi.Token;
  }
});
exports.BizError = BizError;
exports.DeleteMapping = DeleteMapping;
exports.GetMapping = GetMapping;
exports.PatchMapping = PatchMapping;
exports.PostMapping = PostMapping;
exports.Request = Request;
exports.RequestBody = RequestBody;
exports.RequestPathVariable = RequestPathVariable;
exports.RequestQuery = RequestQuery;
exports.Response = Response;
exports.RestController = RestController;
exports.UpdateMapping = UpdateMapping;
exports.___internal = ___internal;
